# -*- coding: utf-8 -*-
#
#% $Id: __init__.py 6998 2009-06-26 08:51:15Z cwilliams $ 
#
#
# Copyright (C) 2002-2007
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import __main__
setattr(__main__,"_meow_verbosity",0);

## ugly hack to get around UGLY FSCKING ARROGNAT (misspelling fully intentional) pyfits-2.3 bug
## (just in case somebody imports pyfits somewhere)
import Kittens.utils
Kittens.utils.import_pyfits();

# import table class
try:
  from pyrap_tables import table,tablecopy,tableexists,tabledelete,addImagingColumns
except:
  try:
    from pyrap.tables import table,tablecopy,tableexists,tabledelete,addImagingColumns
  except:
    print "Failed to import pyrap_tables or pyrap.tables. Please install the pyrap "
    "package from http://code.google.com/p/pyrap/, or from a MeqTrees binary repository "
    "(see http://www.astron.nl/meqwiki/Downloading)"
    raise;

# pyrap likes to display a lot of these, so shut them off
import warnings
warnings.simplefilter('ignore',DeprecationWarning);

# list of optional packages which will be added to the include path
_Packages = [ "Cattery" ];

# list of locations where packages will be searched for
_PackageLocations = [ "~","~/Frameworks/",
  "/usr/local/MeqTrees","/usr/local/lib/MeqTrees","/usr/lib/MeqTrees","/usr/lib64/MeqTrees","/usr/lib32/MeqTrees"
  "/usr/local/meqtrees","/usr/local/lib/meqtrees","/usr/lib/meqtrees","/usr/lib64/meqtrees","/usr/lib32/meqtrees"
  ];

# mapping of package: path. Filled in as we find packages
_packages = {};

import sys
import os
import os.path

def packages ():
  """Returns mapping of available packages to their paths""";
  return _packages;
  # print "Using %s, set the %s_PATH environment variable to override this."%(path,package.upper());

def _tryPackageDir (path,package):
  """Tests if path refers to a valid directory, adds it to system include path if so.
  Marks package as having this path.""";
  if os.path.isdir(path):
    sys.path.insert(0,path);
    # check for version info
    try:
      version = ' '.join(file(os.path.join(path,'version_info')));
    except:
      version = 'no version info';
    # insert into packages
    global _packages;
    _packages[package] = path,version;
    return True;
  return False;

def _setPackagePath (package):
  """Finds the given package, by first looking in $MEQTREES_PACKAGE_PATH, then checking for 
  subdirectories of the standard _PackageLocations list.""";
  # check for explicit MEQTREES_PACKAGE_PATH first
  varname = 'MEQTREES_%s_PATH'%package.upper()
  path = os.environ.get(varname,None);
  if path:
    if not _tryPackageDir(path,package):
      print "Warning: your %s environment variable is set to"%varname;
      print "%s, but this is not a valid directory."%path;
      print "The %s package will not be available."%package;
    return;
  # else look in standard places
  for path in _PackageLocations:
    path = os.path.expanduser(path);
    if _tryPackageDir(os.path.join(path,package),package):
      return;
  # none found
  print "Warning: No %s package found."%package;
  print "If you have %s in a non-standard location, please set the %s environment"%(package,varname);
  print "variable to point to it."

for pkg in _Packages:
  _setPackagePath(pkg);
  
def find_exec (execname):
  import os
  path = os.environ.get('PATH') or os.defpath;
  for dirname in path.split(os.pathsep):
    fname = os.path.join(dirname,execname);
    if os.access(fname,os.R_OK|os.X_OK):
      return fname;
  return None;





#!/usr/bin/python
# -*- coding: utf-8 -*-

#
#% $Id$ 
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import sys
import os.path
import re
import time

time_start = time.time();

from Owlcat import table,tablecopy,tableexists,tabledelete,addImagingColumns
import Owlcat.Console;

progress = Owlcat.Console.Reporter(timestamp=True);

SKIP_COLUMNS = set(("FLAG_CATEGORY","NFRA_AVERAGEDATA","PEELED_DATA","UVMODEL_DATA"));

if __name__ == "__main__":

  # setup some standard command-line option parsing
  #
  from optparse import OptionParser
  parser = OptionParser(usage="""%prog: [options] MS_output MS_in1 MS_in2 ...""",
    description="Concatenates several MSs into one. This is done on a row-by-row basis "
    "with no sanity checking, so it's up to you to insure that the input MSs actually "
    "belong together. All subtables and keywords of the output MS are inherited from "
    "the first input MS, unless the --renumber-spws option is in effect, in which case "
    "the SPECTRAL_WINDOW and DATA_DESCRIPTION tables are also merged (and spwids and "
    "ddids renumbered.)");
  parser.add_option("-a","--append",action="store_true",
                    help="append input MSs to output MS");
  parser.add_option("-f","--force",dest="force",action="store_true",
                    help="proceed without confirmation, and overwrite output MS if it already exists");
  parser.add_option("-s","--renumber-spws",action="store_true",
                    help="treat each MS as a separate spectral window");
  parser.add_option("--renumber-obs",action="store_true",
                    help="renumber OBSERVATION_ID of each MS");
  parser.add_option("-v","--verbose",action="store_true",
                    help="be more verbose");

  (options,msnames) = parser.parse_args();

  if len(msnames) < (2 if options.append else 3):
    parser.error("Insufficient number of arguments. Use '-h' for help.");

  msout = msnames[0];
  msins = msnames[1:];

  if tableexists(msout) and not options.append and not options.force:
    print "Output MS",msout,"already exists. Use the -f switch to overwrite.";
    sys.exit(1);

  if not options.force:
    if options.append:
      print "Will extend merged MS %s with %d input MSs:"%(msout,len(msins));
    else:
      print "Will create merged MS %s from %d input MSs:"%(msout,len(msins));
    for ms in msins:
      print "  ",ms;
    if options.renumber_spws:
      print "Each input MS will be put into a separate spectral window, spws will be renumbered.";
    else:
      print "Spectral windows will not be renumbered.";
    if options.renumber_obs:
      print "OBSERVATION_ID will be incremented for each MS";
    if raw_input("Proceed (y/n)? ").strip().upper()[0] != "Y":
      print "Aborted by user.";
      sys.exit(1);
      
  obsid = 0;

  if not options.append:
    if tableexists(msout) and options.force:
      progress("Deleting existing copy of %s"%msout);
      tabledelete(msout);
    # copy first MS to output as-is
    progress("Copying %s to %s"%(msins[0],msout));
    tablecopy(msins[0],msout,deep=True);
    # when renumbering spectral windows, need to have an addImagingColumns() call
    # unfortunately, this wipes CORRECTED_DATA! I see no way around this (copynorows=True is no help,
    # as then it fails to copy subtables) apart from the ugly: copy first MS deeply,
    # add imaging columns, then re-copy columns
    if options.renumber_spws:
      addImagingColumns(msout);
    # otherwise, first MS is already copied, so remove it from list
    # (but reset OBSERVATION_ID)
    else:
      if options.renumber_obs:
        tab = table(msout,readonly=False);
        obs = tab.getcol("OBSERVATION_ID");
        obs[:] = obsid;
        obsid += 1;
        tab.putcol("OBSERVATION_ID",obs);
        tab.close();
      msins = msins[1:];

  # open output for writing
  tab0 = table(msout,readonly=False);

  # get the number of DDIDs and SPWIDs
  ddid_offsets = [0]*len(msins);
  if options.renumber_spws:
    ddid_tab0 = table(tab0.getkeyword("DATA_DESCRIPTION"),readonly=False);
    spw_tab0 = table(tab0.getkeyword("SPECTRAL_WINDOW"),readonly=False);
    # now, somehow lwimager will still try to re-init MODEL_DATA and CORRECTED_DATA when dealing with merged
    # MSs. The only way around it that I can see is to fill up the ddid and spw tabs first
    progress("Filling new DATA_DESCRIPTION and SPECTRAL_WINDOW table");
    for num_ms,ms in enumerate(msins[1:]):
      tab = table(ms);
      ddid_tab = table(tab.getkeyword("DATA_DESCRIPTION"),readonly=False);
      nr0 = ddid_tab0.nrows();
      ddid_offsets[num_ms+1] = nr0;
      ddid_tab0.addrows(ddid_tab.nrows());
      for col in ddid_tab.colnames():
        data = ddid_tab.getcol(col);
        if col == "SPECTRAL_WINDOW_ID":
          data += spw_tab0.nrows();
        ddid_tab0.putcol(col,data,nr0);
      # append content of the SPECTRAL_WINDOW
      spw_tab = table(tab.getkeyword("SPECTRAL_WINDOW"),readonly=False);
      nr0 = spw_tab0.nrows();
      spw_tab0.addrows(spw_tab.nrows());
      for col in spw_tab.colnames():
        data = spw_tab.getcol(col);
        spw_tab0.putcol(col,data,nr0);
    num_spws,num_ddids = spw_tab0.nrows(),ddid_tab0.nrows();
    ddid_tab0.close();
    spw_tab0.close();
    tab0.close();
    addImagingColumns(msout);
    tab0 = table(msout,readonly=False);

  overprint = progress if options.verbose else progress.overprint;

  for num_ms,msname in enumerate(msins):
    tab = table(msname);
    nr0 = tab0.nrows();
    # in spw renumbering mode, because of the ugliness explained above,
    # we need to re-copy the data from the first MS
    if options.renumber_spws and not num_ms:
      nr0 = 0;
    # else, insert more rows to accommodate added MS
    else:
      tab0.addrows(tab.nrows());
    progress("Current row count: %d"%nr0);
    progress("Merging in %d rows from %s"%(tab.nrows(),msname));
    for col in tab.colnames():
      if col not in SKIP_COLUMNS:
        overprint("Reading column %s"%col);
        try:
          data = tab.getcol(col);
        except:
          print "WARNING: MS %s does not appear to contain a valid %s column"%(msname,col);
          print "   This is not necessarily fatal, proceeding with the merge anyway."
          continue;
        if options.renumber_obs and col == "OBSERVATION_ID":
          data[:] = obsid;
          obsid += 1;
        # if renumbering DDIDs, increment the DDID column by the # of rows in the DDID table --
        # this ensures uniqueness of DDIDs
        if options.renumber_spws and col == "DATA_DESC_ID":
          progress("Adding %d to DATA_DESC_ID"%ddid_offsets[num_ms]);
          data += ddid_offsets[num_ms];
        overprint("Writing column %s, shape %s"%(col,data.shape));
        tab0.putcol(col,data,nr0);


  overprint("Closing output MS %s\n"%msout);
  nr0 = tab0.nrows();
  progress("Wrote %d rows to output MS"%nr0);
  if options.renumber_spws:
    progress("Output MS contains %d spectral windows and %d DDIDs"%(num_spws,num_ddids));
  tab0.close();



#!/usr/bin/python
# -*- coding: utf-8 -*-

#
#% $Id$
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import sys
import os.path
import re
import time
import math
import numpy
import Owlcat

ROWCHUNK = 50000;

time_start = time.time();

try:
  from pyrap_tables import table,tablecopy,tableexists,tabledelete
except:
  from pyrap.tables import table,tablecopy,tableexists,tabledelete

def timestamp (format="%H:%M:%S:"):
  return time.strftime(format,time.gmtime(time.time()-time_start));

def progress (message,newline=True):
  sys.stdout.write("%s %-60s%s"%(timestamp(),message,'\n' if newline else '\r'));
  sys.stdout.flush();

if __name__ == "__main__":

  # setup some standard command-line option parsing
  #
  from optparse import OptionParser
  parser = OptionParser(usage="""%prog: [options] MS""",
      description="""Finds redundant baselines and sets the WEIGHT (or
IMAGING_WEIGHT) column of the MS, weighting the redundant baselines as 1/n. For a redundant array such as WSRT, this considerably reduces sidelobes in the image.""");
  #parser.add_option("-o","--output",dest="output",type="string",
                    #help="name of output FITS file");
  #parser.add_option("-z","--zoom",dest="output_quad",type="string",
                    #help="name of zoomed output FITS file");
  parser.add_option("-t","--tolerance",dest="tolerance",type="float",
                    help="How close (in meters) two baselines need to be to each other to be considered redundant (default .1)");
  parser.add_option("-w","--weight",action="store_true",
                    help="If this option is given, the IMAGING_WEIGHT column will be "
                    "scaled by 1/n (where n is the number of redundant baselines). The "
                    "default behaviour is to simply set the WEIGHT column to 1/n.")
  parser.add_option("-I","--ifrs",dest="ifrs",type="string",
                    help="subset of interferometers to use.");
  parser.add_option("-s","--select",dest="select",action="store",
                    help="additional TaQL selection string. Note that redundant baselines are counted only within the subset "
                         "given by the --ifrs and --select options.");
  parser.add_option("-l","--list",action="store_true",
                    help="list all baselines and exit.");
  parser.add_option("-r","--reset",action="store_true",
                    help="reset WEIGHT column to unity. All other options are then ignored.");
  parser.set_defaults(tolerance=.1,select="",ifrs="");

  (options,msnames) = parser.parse_args();

  if len(msnames) != 1:
    parser.error("MS name not supplied.");

  msname = msnames[0];
  ms = table(msname,readonly=False);
  ncorr = ms.getcol('DATA',0,1).shape[2];

  if options.reset:
    # apply weights
    nrows = ms.nrows();
    for row0 in range(0,nrows,ROWCHUNK):
      progress("Resetting weights, row %d of %d"%(row0,nrows),newline=False);
      row1 = min(nrows,row0+ROWCHUNK);
      w = numpy.ones((row1-row0,ncorr),float);
      ms.putcol('WEIGHT',w,row0,row1-row0);
    progress("Weights reset to unity.");
  else:
    taqls = [];
    # get IFR set
    import Meow.IfrSet
    ifrset = Meow.IfrSet.from_ms(ms);
    if options.ifrs:
      ifrset = ifrset.subset(options.ifrs);
      taqls.append(ifrset.taql_string());

    if options.select:
      taqls.append(options.select);

    if taqls:
      select = "( " + " ) &&( ".join(taqls) + " )";
      progress("Applying TaQL selection %s"%select,newline=True);
      ms = ms.query(select);
    progress("Looking for redundant baselines",newline=True);
    ant1 = ms.getcol('ANTENNA1');
    ant2 = ms.getcol('ANTENNA2');

    IFRS = sorted(set([ (p,q) for p,q in zip(ant1,ant2) ]));
    print "%d baselines"%len(IFRS);
    groups = [];

    for i,(p,q) in enumerate(IFRS):
      bl = ifrset.baseline_vector(p,q);
      # see if this baseline is within the tolerance of a previous group's baseline
      for ig,(bl0,members) in enumerate(groups):
        if abs(bl-bl0).max() < options.tolerance:
          members.append((p,q));
          break;
      # if none, start a new group
      else:
        members = [(p,q)];
        ig = len(groups);
        groups.append([]);
      # update baseline (as mean baseline of group)
      length = reduce(lambda x,y:x+y,[ifrset.baseline_vector(*ifr) for ifr in members])/len(members);
      groups[ig] = (length,members);

    # convert to list of length,members, and sort by length
    groups = [ (math.sqrt((baseline**2).sum()),members) for baseline,members in groups ];
    groups.sort();

    if options.list:
      baselist = [ "%dm (%s)"%(round(length)," ".join(["%s-%s"%(p,q) for p,q in mem]))
          for length,mem in groups ];
      print "Found %d non-redundant baselines:"%len(baselist),", ".join(baselist);
      sys.exit(0);

    # make a dictionary of per-IFR weights
    have_redundancy = False;
    weight = dict([((p,q),1.) for p,q in IFRS]);
    for baseline,members in groups:
      if len(members) > 1:
        print "Baseline %dm, %d ifrs: %s"%(round(baseline),len(members),
          " ".join(["%d-%d"%(p,q) for p,q in members]));
        have_redundancy = True;
        for p,q in members:
          weight[p,q] = 1.0/len(members);

    if not have_redundancy:
      print "No redundant baselines found, nothing to do."
      sys.exit(0);

    # apply weights
    nrows = ms.nrows();
    for row0 in range(0,nrows,ROWCHUNK):
      progress("Applying weights, row %d of %d"%(row0,nrows),newline=False);
      row1 = min(nrows,row0+ROWCHUNK);
      if not options.weight:
        w = numpy.zeros((row1-row0,ncorr),float);
        for i in range(row0,row1):
          w[i-row0,:] = weight[ant1[i],ant2[i]];
        ms.putcol('WEIGHT',w,row0,row1-row0);
      else:
        w = ms.getcol('IMAGING_WEIGHT',row0,row1-row0);
        for i in range(row0,row1):
          w[i-row0,:] *= weight[ant1[i],ant2[i]];
        ms.putcol('IMAGING_WEIGHT',w,row0,row1-row0);

  progress("Closing MS.");
  ms.close();



# -*- coding: utf-8 -*-

import os
import sys
import time
#import curses

# This stuff doesn't work when running the script in the background!
# Dunno why, so disabling this for 1.2

## Get width of terminal
## I couldn't be bothered to get newline mode working properly when curses is active,
## so I used curses.wrapper() do init curses, get the width, then close curses.
#def get_width (scr):
  #global _width;
  #_width = scr.getmaxyx()[1] - 1;

## Since curses can fall over when invoked via ssh non-interactively (curse you curses!),
## protect this with an exception
#try:
  #curses.wrapper(get_width);
#except:
  #_width = 80;

try:
  _height,_width = map(int,os.popen('stty size', 'r').read().split())
except:
  _width = 132;

def timestamp (time_start,format="%H:%M:%S:"):
  return time.strftime(format,time.gmtime(time.time()-time_start));

class Reporter (object):
  """A Reporter is used to make progress reports on the console, with optional timestamps."""
  def __init__ (self,timestamp=False):
    self.time_start = timestamp and time.time();

  def overprint (self,message):
    return self.pprint(message+"\r");

  def pprint (self,message,newline=True):
    # print message
    if self.time_start:
      message = "%s %s"%(timestamp(self.time_start),message); 
    if message[-1] == "\r":
      endline = "\r";
    else:
      endline = "\n\r";
    # cut message at width of terminal, and pad with spaces
    sys.stdout.write("%-*.*s"%(_width,_width,message.strip()));
    sys.stdout.write(endline);
    sys.stdout.flush();

  def __call__ (self,*args):
    self.pprint(" ".join(args));
    
  
  


# -*- coding: utf-8 -*-


import cPickle
import gzip


class TableDump (object):
  # initializes a table dump file
  # format:
  #   header dict with version=2, 
  #   followed by colname,keywords,nchunks(int),chunk1,...,chunkN for each column
  def __init__ (self,filename,version=1,compress=True,write=False):
    if compress:
      self.stream = gzip.GzipFile(filename,"w" if write else "r");
    else:
      self.stream = file(filename,"wb" if write else "rb");
    if write:
      cPickle.dump(dict(format='Owlcat MS column dump',version=version),self.stream);
    else:
      self.header = cPickle.load(self.stream);
      
  def __del__ (self):
    self.close();
    
  def close (self):
    self.stream.close();

  def dump_column (self,tab,colname,verbose=False):
    """Exports the specified column of the table""" 
    nrows = tab.nrows();
    kws = tab.getcolkeywords(colname);
    # dump in <.5Gb chunks, so figure out size of one row
    row0 = tab.getcol(colname,0,1);
    maxrows = (2**27)//(row0.size*row0.itemsize);
    rng = range(0,nrows,maxrows);
    # dump name,keywords,num_chunks
    for obj in colname,kws,len(rng):
      cPickle.dump(obj,self.stream);
    # dump chunks one by one
    for i0 in rng:
      cPickle.dump(tab.getcol(colname,i0,maxrows),self.stream);
    if verbose:
      print "  %s: %d chunk(s)%s%s"%(colname,len(rng),
        ", array column shape %s"%"x".join(map(str,row0.shape[1:])) if row0.ndim>2 else "scalar column",
        ", %d keywords"%len(kws) if kws else "");

  def load (self,tab,verbose):
    """Loads all columns from the dump file""" 
    colnames = set(tab.colnames());
    while True:
      try:
        colname = cPickle.load(self.stream);
      except EOFError:
        break;
      if not colname:
        break;
      if colname in colnames:
        kws,nchunks = cPickle.load(self.stream),cPickle.load(self.stream);
        if kws is not None:
          tab.putcolkeywords(colname,kws);
        row0 = 0;
        for i in range(nchunks):
          col = cPickle.load(self.stream);
#          print col.shape,row0;
          tab.putcol(colname,col,row0,col.shape[0]);
          row0 += col.shape[0];
        if verbose:
          print "  %s: %d chunks%s"%(colname,nchunks,", %d keywords"%len(kws) if kws else "");
